# -*- coding: utf-8 -*-
"""
app.main.views
~~~~~~~~~~~~~~

This module implements the main blueprint.

:copyright: (c) 2017 European Spallation Source ERIC
:license: BSD 2-Clause, see LICENSE for more details.

"""
import sqlalchemy as sa
from flask import (Blueprint, render_template, jsonify,
                   redirect, url_for, request, flash)
from flask_login import login_required
from .forms import QRCodeForm
from ..extensions import db
from ..models import Action, Manufacturer, Model, Location, Status, Item
from ..decorators import login_groups_accepted
from .. import utils

bp = Blueprint('main', __name__)


# Declare custom error handlers for all views
@bp.app_errorhandler(404)
def not_found_error(error):
    return render_template('404.html'), 404


@bp.app_errorhandler(500)
def internal_error(error):
    return render_template('500.html'), 500


@bp.app_errorhandler(utils.CSEntryError)
def handle_csentry_error(error):
    response = jsonify(error.to_dict())
    response.status_code = error.status_code
    return response


@bp.route('/_retrieve_items')
@login_required
def retrieve_items():
    items = Item.query.order_by(Item._created)
    data = [[item.id,
             item.ics_id,
             utils.format_field(item._created),
             utils.format_field(item._updated),
             item.serial_number,
             utils.format_field(item.manufacturer),
             utils.format_field(item.model),
             utils.format_field(item.location),
             utils.format_field(item.status),
             utils.format_field(item.parent),
             ] for item in items]
    return jsonify(data=data)


@bp.route('/')
@login_required
def index():
    return render_template('index.html')


@bp.route('/view/<ics_id>')
@login_required
def view_item(ics_id):
    item = Item.query.filter_by(ics_id=ics_id).first_or_404()
    return render_template('view_item.html', item=item.to_dict(long=True))


@bp.route('/qrcodes')
@login_required
def qrcodes():
    codes = {}
    for model in (Action, Manufacturer, Model, Location, Status):
        items = db.session.query(model).order_by(model.name)
        images = [{'name': item.name,
                   'data': utils.image_to_base64(item.image())}
                  for item in items]
        codes[model.__name__] = images
    return render_template('qrcodes.html', codes=codes)


@bp.route('/qrcodes/create', methods=('GET', 'POST'))
@login_groups_accepted('admin', 'create')
def create_qrcodes():
    kind = request.args.get('kind', 'Manufacturer')
    form = QRCodeForm(request.form, kind=kind)
    if form.validate_on_submit():
        model = globals()[form.kind.data]
        new_model = model(name=form.name.data)
        db.session.add(new_model)
        try:
            db.session.commit()
        except sa.exc.IntegrityError as e:
            db.session.rollback()
            flash(f'{form.name.data} already exists! Item not created.', 'error')
        return redirect(url_for('main.create_qrcodes', kind=form.kind.data))
    return render_template('create_qrcodes.html', form=form)


@bp.route('/_retrieve_qrcodes_name/<kind>')
@login_required
def retrieve_qrcodes_name(kind):
    try:
        model = globals()[kind]
    except KeyError:
        raise utils.CSEntryError(f"Unknown model '{kind}'", status_code=422)
    items = db.session.query(model).order_by(model.name)
    data = [[item.name] for item in items]
    return jsonify(data=data)
