# -*- coding: utf-8 -*-
"""
tests.functional.test_api
~~~~~~~~~~~~~~~~~~~~~~~~~

This module defines API tests.

:copyright: (c) 2017 European Spallation Source ERIC
:license: BSD 2-Clause, see LICENSE for more details.

"""
import json
import pytest
from app import models


ENDPOINT_MODEL = {
    'actions': models.Action,
    'manufacturers': models.Manufacturer,
    'models': models.Model,
    'locations': models.Location,
    'status': models.Status,
    'items': models.Item,
    'networks': models.Network,
}
GENERIC_GET_ENDPOINTS = [key for key in ENDPOINT_MODEL.keys() if key not in ('items', 'networks')]
GENERIC_CREATE_ENDPOINTS = [key for key in ENDPOINT_MODEL.keys() if key not in ('items', 'actions', 'networks')]
CREATE_AUTH_ENDPOINTS = [key for key in ENDPOINT_MODEL.keys() if key != 'actions']


def get(client, url, token=None):
    response = client.get(
        url,
        headers={'Content-Type': 'application/json',
                 'Authorization': f'Bearer {token}'},
    )
    if response.headers['Content-Type'] == 'application/json':
        response.json = json.loads(response.data)
    return response


def post(client, url, data, token=None):
    headers = {'Content-Type': 'application/json'}
    if token is not None:
        headers['Authorization'] = f'Bearer {token}'
    response = client.post(url, data=json.dumps(data), headers=headers)
    if response.headers['Content-Type'] == 'application/json':
        response.json = json.loads(response.data)
    return response


def patch(client, url, data, token=None):
    headers = {'Content-Type': 'application/json'}
    if token is not None:
        headers['Authorization'] = f'Bearer {token}'
    response = client.patch(url, data=json.dumps(data), headers=headers)
    if response.headers['Content-Type'] == 'application/json':
        response.json = json.loads(response.data)
    return response


def login(client, username, password):
    data = {
        'username': username,
        'password': password
    }
    return post(client, '/api/login', data)


def get_token(client, username, password):
    response = login(client, username, password)
    return response.json['access_token']


@pytest.fixture()
def readonly_token(client):
    return get_token(client, 'user_ro', 'userro')


@pytest.fixture()
def user_token(client):
    return get_token(client, 'user_rw', 'userrw')


@pytest.fixture()
def admin_token(client):
    return get_token(client, 'admin', 'adminpasswd')


def check_response_message(response, msg, status_code=400):
    assert response.status_code == status_code
    try:
        data = response.json
    except AttributeError:
        data = json.loads(response.data)
    try:
        message = data['message']
    except KeyError:
        # flask-jwt-extended is using "msg" instead of "message"
        # in its default callbacks
        message = data['msg']
    assert message == msg


def check_names(response, names):
    response_names = set(item['name'] for item in response.json)
    assert set(names) == response_names


def check_input_is_subset_of_response(response, inputs):
    # Sort the response by id to match the inputs order
    response_elts = sorted(response.json, key=lambda d: d['id'])
    for d1, d2 in zip(inputs, response_elts):
        assert set(d1.items()).issubset(set(d2.items()))


def test_login(client):
    response = client.post('/api/login')
    check_response_message(response, 'Body should be a JSON object')
    response = post(client, '/api/login', data={'username': 'foo', 'passwd': ''})
    check_response_message(response, 'Missing mandatory field (username or password)', 422)
    response = login(client, 'foo', 'invalid')
    check_response_message(response, 'Invalid credentials', 401)
    response = login(client, 'user_ro', 'userro')
    assert response.status_code == 200
    assert 'access_token' in response.json


@pytest.mark.parametrize('endpoint', GENERIC_GET_ENDPOINTS)
def test_get_generic_model(endpoint, session, client, readonly_token):
    model = ENDPOINT_MODEL[endpoint]
    names = ('Foo', 'Bar', 'Alice')
    for name in names:
        session.add(model(name=name))
    session.commit()
    response = client.get(f'/api/{endpoint}')
    check_response_message(response, 'Missing Authorization Header', 401)
    response = get(client, f'/api/{endpoint}', 'xxxxxxxxx')
    check_response_message(response, 'Not enough segments', 422)
    response = get(client, f'/api/{endpoint}', readonly_token)
    check_names(response, names)
    response = get(client, f'/api/{endpoint}?qrcode=true', readonly_token)
    check_names(response, names)
    for item in response.json:
        assert 'qrcode' in item
    response = get(client, f'/api/{endpoint}?qrcode=false', readonly_token)
    check_names(response, names)
    for item in response.json:
        assert 'qrcode' not in item


@pytest.mark.parametrize('endpoint', CREATE_AUTH_ENDPOINTS)
def test_create_model_auth_fail(endpoint, client, readonly_token):
    response = client.post(f'/api/{endpoint}')
    check_response_message(response, 'Missing Authorization Header', 401)
    response = post(client, f'/api/{endpoint}', data={}, token='xxxxxxxxx')
    check_response_message(response, 'Not enough segments', 422)
    response = post(client, f'/api/{endpoint}', data={}, token=readonly_token)
    check_response_message(response, "User doesn't have the required group", 403)
    model = ENDPOINT_MODEL[endpoint]
    assert model.query.count() == 0


@pytest.mark.parametrize('endpoint', GENERIC_CREATE_ENDPOINTS)
def test_create_generic_model(endpoint, client, user_token):
    response = post(client, f'/api/{endpoint}', data={}, token=user_token)
    check_response_message(response, "Missing mandatory field 'name'", 422)
    data = {'name': 'Foo'}
    response = post(client, f'/api/{endpoint}', data=data, token=user_token)
    assert response.status_code == 201
    assert {'id', 'name'} <= set(response.json.keys())
    assert response.json['name'] == 'Foo'
    response = post(client, f'/api/{endpoint}', data=data, token=user_token)
    check_response_message(response, 'IntegrityError', 409)
    response = post(client, f'/api/{endpoint}', data={'name': 'foo'}, token=user_token)
    check_response_message(response, 'IntegrityError', 409)
    response = post(client, f'/api/{endpoint}', data={'name': 'FOO'}, token=user_token)
    check_response_message(response, 'IntegrityError', 409)
    data = {'name': 'Bar'}
    response = post(client, f'/api/{endpoint}', data=data, token=user_token)
    assert response.status_code == 201
    model = ENDPOINT_MODEL[endpoint]
    assert model.query.count() == 2
    response = get(client, f'/api/{endpoint}', user_token)
    check_names(response, ('Foo', 'Bar'))


def test_create_item(client, user_token):
    # check that serial_number is mandatory
    response = post(client, '/api/items', data={}, token=user_token)
    check_response_message(response, "Missing mandatory field 'serial_number'", 422)

    # check create with only serial_number
    data = {'serial_number': '123456'}
    response = post(client, '/api/items', data=data, token=user_token)
    assert response.status_code == 201
    assert {'id', 'ics_id', 'serial_number', 'manufacturer', 'model',
            'location', 'status', 'updated', 'created', 'parent'} == set(response.json.keys())
    assert response.json['serial_number'] == '123456'

    # Check that serial_number doesn't have to be unique
    response = post(client, '/api/items', data=data, token=user_token)
    assert response.status_code == 201

    # check that ics_id shall be unique
    data2 = {'serial_number': '456789', 'ics_id': 'AAA001'}
    response = post(client, '/api/items', data=data2, token=user_token)
    assert response.status_code == 201
    response = post(client, '/api/items', data=data2, token=user_token)
    check_response_message(response, 'IntegrityError', 409)

    # check all items that were created
    assert models.Item.query.count() == 3
    response = get(client, '/api/items', user_token)
    check_input_is_subset_of_response(response, (data, data, data2))


def test_create_item_invalid_ics_id(client, user_token):
    for ics_id in ('foo', 'AAB1234', 'AZ02', 'WS007', 'AAA01'):
        data = {'serial_number': '123456', 'ics_id': ics_id}
        response = post(client, '/api/items', data=data, token=user_token)
        check_response_message(response, 'ICS id shall match [A-Z]{3}[0-9]{3}', 422)


def test_get_item_fail(client, session, readonly_token):
    response = get(client, '/api/items/50', token=readonly_token)
    check_response_message(response, "Item id '50' not found", 404)
    response = get(client, '/api/items/bar', token=readonly_token)
    check_response_message(response, "Item id 'bar' not found", 404)


def test_get_item(client, session, readonly_token):
    # Create some items
    session.add(models.Status(name='Stock'))
    session.flush()
    item1 = models.Item(serial_number='123456')
    item2 = models.Item(serial_number='234567', ics_id='AAA001', status='Stock')
    for item in (item1, item2):
        session.add(item)
    session.commit()

    # we can get items by id...
    response = get(client, f'/api/items/{item1.id}', token=readonly_token)
    assert response.status_code == 200
    assert response.json['id'] == item1.id
    assert response.json['serial_number'] == item1.serial_number
    # ...or ics_id
    response = get(client, f'/api/items/{item2.ics_id}', token=readonly_token)
    assert response.status_code == 200
    assert response.json['id'] == item2.id
    assert response.json['ics_id'] == item2.ics_id
    assert response.json['serial_number'] == item2.serial_number
    assert response.json['status'] == str(item2.status)


def test_patch_item_auth_fail(client, session, readonly_token):
    response = client.patch('/api/items/50')
    check_response_message(response, 'Missing Authorization Header', 401)
    response = patch(client, '/api/items/50', data={}, token='xxxxxxxxx')
    check_response_message(response, 'Not enough segments', 422)
    response = patch(client, '/api/items/50', data={}, token=readonly_token)
    check_response_message(response, "User doesn't have the required group", 403)


def test_patch_item_fail(client, session, user_token):
    response = patch(client, '/api/items/50', data={}, token=user_token)
    check_response_message(response, 'At least one field is required', 422)
    data = {'location': 'ESS', 'foo': 'bar'}
    response = patch(client, '/api/items/50', data=data, token=user_token)
    check_response_message(response, "Invalid field 'foo'", 422)
    data = {'location': 'ESS'}
    response = patch(client, '/api/items/50', data=data, token=user_token)
    check_response_message(response, "Item id '50' not found", 404)
    response = patch(client, '/api/items/bar', data=data, token=user_token)
    check_response_message(response, "Item id 'bar' not found", 404)

    # Create an item
    item1 = models.Item(serial_number='234567', ics_id='AAA001')
    session.add(item1)
    session.commit()

    # check that we can't change the serial_number or ics_id
    response = patch(client, f'/api/items/{item1.id}', data={'serial_number': '12345'}, token=user_token)
    check_response_message(response, "Invalid field 'serial_number'", 422)
    response = patch(client, f'/api/items/{item1.id}', data={'ics_id': 'AAA002'}, token=user_token)
    check_response_message(response, "'ics_id' can't be changed", 422)


def test_patch_item(client, session, user_token):
    # Create some items
    session.add(models.Status(name='Stock'))
    session.add(models.Status(name='In service'))
    session.flush()
    item1 = models.Item(serial_number='123456')
    item2 = models.Item(serial_number='234567', ics_id='AAA001', status='Stock')
    for item in (item1, item2):
        session.add(item)
    session.commit()

    # we can patch items by id...
    data = {'ics_id': 'AAA004'}
    response = patch(client, f'/api/items/{item1.id}', data=data, token=user_token)
    assert response.status_code == 200
    assert response.json['id'] == item1.id
    assert response.json['serial_number'] == item1.serial_number
    assert response.json['ics_id'] == data['ics_id']
    # ...or ics_id
    data = {'status': 'In service'}
    response = patch(client, f'/api/items/{item2.ics_id}', data=data, token=user_token)
    assert response.status_code == 200
    assert response.json['id'] == item2.id
    assert response.json['ics_id'] == item2.ics_id
    assert response.json['serial_number'] == item2.serial_number
    assert response.json['status'] == data['status']


def test_patch_item_integrity_error(client, session, user_token):
    # Create some items
    item1 = models.Item(serial_number='123456', ics_id='AAA001')
    item2 = models.Item(serial_number='234567')
    for item in (item1, item2):
        session.add(item)
    session.commit()
    data = {'ics_id': 'AAA001'}
    response = patch(client, f'/api/items/{item2.id}', data=data, token=user_token)
    check_response_message(response, 'IntegrityError', 409)


def test_patch_item_parent(client, session, user_token):
    # Create some items
    session.add(models.Manufacturer(name='HP'))
    session.add(models.Manufacturer(name='Dell'))
    session.add(models.Status(name='Stock'))
    session.add(models.Status(name='In service'))
    session.add(models.Location(name='ESS'))
    session.add(models.Location(name='ICS lab'))
    session.flush()
    item1 = models.Item(serial_number='123456', ics_id='AAA001', location='ICS lab', status='In service',
                        manufacturer='Dell')
    item2 = models.Item(serial_number='234567', ics_id='AAA002', status='Stock')
    item3 = models.Item(serial_number='345678', ics_id='AAA003', status='Stock', manufacturer='Dell')
    for item in (item1, item2, item3):
        session.add(item)
    session.flush()
    item3.parent_id = item1.id
    session.commit()

    # set parent changes the status and location
    data1 = {'parent': item1.ics_id}
    response = patch(client, f'/api/items/{item2.ics_id}', data=data1, token=user_token)
    assert response.status_code == 200
    assert response.json['id'] == item2.id
    assert response.json['ics_id'] == item2.ics_id
    assert response.json['serial_number'] == item2.serial_number
    assert response.json['parent'] == item1.ics_id
    assert response.json['status'] == str(item1.status)
    assert response.json['location'] == str(item1.location)

    # updating a parent, modifies the status and location of all children
    # check location
    data2 = {'location': 'ESS'}
    response = patch(client, f'/api/items/{item1.ics_id}', data=data2, token=user_token)
    assert response.status_code == 200
    assert response.json['id'] == item1.id
    assert response.json['ics_id'] == item1.ics_id
    assert response.json['serial_number'] == item1.serial_number
    assert response.json['status'] == str(item1.status)
    assert response.json['location'] == data2['location']
    for ics_id in ('AAA002', 'AAA003'):
        response = get(client, f'/api/items/{ics_id}', token=user_token)
        assert response.json['location'] == data2['location']
        assert response.json['status'] == 'In service'
    # check status
    data3 = {'status': 'Stock'}
    response = patch(client, f'/api/items/{item1.ics_id}', data=data3, token=user_token)
    assert response.status_code == 200
    assert response.json['status'] == data3['status']
    for ics_id in ('AAA002', 'AAA003'):
        response = get(client, f'/api/items/{ics_id}', token=user_token)
        assert response.json['location'] == data2['location']
        assert response.json['status'] == data3['status']

    # manufacturer has no impact on children
    data4 = {'manufacturer': 'HP'}
    response = patch(client, f'/api/items/{item1.ics_id}', data=data4, token=user_token)
    assert response.status_code == 200
    assert response.json['manufacturer'] == 'HP'
    # Manufacturer didn't change on children
    response = get(client, f'/api/items/{item2.ics_id}', token=user_token)
    assert response.json['manufacturer'] is None
    response = get(client, f'/api/items/{item3.ics_id}', token=user_token)
    assert response.json['manufacturer'] == 'Dell'


def test_get_items(client, session, readonly_token):
    # Create some items
    session.add(models.Manufacturer(name='Dell'))
    session.add(models.Status(name='Stock'))
    session.add(models.Status(name='In service'))
    session.add(models.Location(name='ESS'))
    session.add(models.Location(name='ICS lab'))
    session.flush()
    item1 = models.Item(serial_number='123456', ics_id='AAA001', location='ESS', status='In service',
                        manufacturer='Dell')
    item2 = models.Item(serial_number='234567', ics_id='AAA002', status='Stock')
    item3 = models.Item(serial_number='345678', ics_id='AAA003', status='Stock', manufacturer='Dell')
    for item in (item1, item2, item3):
        session.add(item)
    session.commit()

    response = get(client, '/api/items', token=readonly_token)
    assert response.status_code == 200
    assert len(response.json) == 3
    check_input_is_subset_of_response(response, (item1.to_dict(), item2.to_dict(), item3.to_dict()))

    # test filtering
    response = get(client, '/api/items?serial_number=234567', token=readonly_token)
    assert response.status_code == 200
    assert len(response.json) == 1
    check_input_is_subset_of_response(response, (item2.to_dict(),))
    # filtering on location_id works but not location (might want to change that)
    response = get(client, f'/api/items?location_id={item1.location_id}', token=readonly_token)
    assert response.status_code == 200
    assert len(response.json) == 1
    check_input_is_subset_of_response(response, (item1.to_dict(),))
    response = get(client, '/api/items?location=ESS', token=readonly_token)
    check_response_message(response, 'Invalid query arguments', 422)
    # using an unknown key raises a 422
    response = get(client, '/api/items?foo=bar', token=readonly_token)
    check_response_message(response, 'Invalid query arguments', 422)


def test_get_networks(client, session, readonly_token):
    # Create some networks
    location = models.Location(name='G02')
    session.add(location)
    session.flush()
    network1 = models.Network(prefix='172.16.1.0/24', first='172.16.1.1', last='172.16.1.254', label='network1')
    network2 = models.Network(prefix='172.16.20.0/22', first='172.16.20.11', last='172.16.20.250')
    network3 = models.Network(prefix='172.16.5.0/24', first='172.16.5.10', last='172.16.5.254', location_id=location.id)
    for network in (network1, network2, network3):
        session.add(network)
    session.commit()

    response = get(client, '/api/networks', token=readonly_token)
    assert response.status_code == 200
    assert len(response.json) == 3
    check_input_is_subset_of_response(response, (network1.to_dict(), network2.to_dict(), network3.to_dict()))

    # test filtering by location_id
    response = get(client, f'/api/networks?location_id={location.id}', token=readonly_token)
    assert response.status_code == 200
    assert len(response.json) == 1
    check_input_is_subset_of_response(response, (network3.to_dict(),))

    # test filtering by prefix
    response = get(client, '/api/networks?prefix=172.16.20.0/22', token=readonly_token)
    assert response.status_code == 200
    assert len(response.json) == 1
    check_input_is_subset_of_response(response, (network2.to_dict(),))


def test_create_network_auth_fail(client, session, user_token):
    # admin is required to create networks
    response = post(client, '/api/networks', data={}, token=user_token)
    check_response_message(response, "User doesn't have the required group", 403)


def test_create_network(client, session, admin_token):
    location = models.Location(name='G02')
    session.add(location)
    session.commit()
    # check that prefix, first and last are mandatory
    response = post(client, '/api/networks', data={}, token=admin_token)
    check_response_message(response, "Missing mandatory field 'prefix'", 422)
    response = post(client, '/api/networks', data={'first': '172.16.1.10', 'last': '172.16.1.250'}, token=admin_token)
    check_response_message(response, "Missing mandatory field 'prefix'", 422)
    response = post(client, '/api/networks', data={'prefix': '172.16.1.0/24'}, token=admin_token)
    check_response_message(response, "Missing mandatory field 'first'", 422)
    response = post(client, '/api/networks', data={'prefix': '172.16.1.0/24', 'first': '172.16.1.10'}, token=admin_token)
    check_response_message(response, "Missing mandatory field 'last'", 422)

    data = {'prefix': '172.16.1.0/24',
            'first': '172.16.1.10',
            'last': '172.16.1.250'}
    response = post(client, '/api/networks', data=data, token=admin_token)
    assert response.status_code == 201
    assert {'id', 'prefix', 'first', 'last', 'label', 'vlanid', 'gateway', 'location'} == set(response.json.keys())
    assert response.json['prefix'] == '172.16.1.0/24'
    assert response.json['first'] == '172.16.1.10'
    assert response.json['last'] == '172.16.1.250'

    # Check that prefix shall be unique
    response = post(client, '/api/networks', data=data, token=admin_token)
    check_response_message(response, 'IntegrityError', 409)

    # Check that all parameters can be passed
    data2 = {'prefix': '172.16.5.0/24',
             'first': '172.16.5.11',
             'last': '172.16.5.250',
             'label': 'G02',
             'gateway': '172.16.5.10',
             'vlanid': 1601,
             'location_id': location.id}
    response = post(client, '/api/networks', data=data2, token=admin_token)
    assert response.status_code == 201
    assert response.json['location'] == location.name

    # check all items that were created
    assert models.Network.query.count() == 2


def test_create_network_constraint_fail(client, session, admin_token):
    # first not in prefix
    data = {'prefix': '172.16.1.0/24',
            'first': '172.16.2.10',
            'last': '172.16.1.250'}
    response = post(client, '/api/networks', data=data, token=admin_token)
    check_response_message(response, 'IntegrityError', 409)
    # last not in prefix
    data = {'prefix': '172.16.1.0/24',
            'first': '172.16.1.10',
            'last': '172.16.5.250'}
    response = post(client, '/api/networks', data=data, token=admin_token)
    check_response_message(response, 'IntegrityError', 409)
    # first > last
    data = {'prefix': '172.16.1.0/24',
            'first': '172.16.1.10',
            'last': '172.16.1.9'}
    response = post(client, '/api/networks', data=data, token=admin_token)
    check_response_message(response, 'IntegrityError', 409)
